{-# OPTIONS_GHC -fglasgow-exts -fallow-overlapping-instances #-}
module Hawl.MonadWeb (Web, runWeb, runWebSession,
                      URL, URI,
                      BaseConfig(..), getConfig, fromConfig, withConfig,
                      configBaseURI, configBaseURL,
                      loadSession, startSession, loadOrStartSession,
                      endSession,
                      isLoggedIn,
                      module Hawl.HaskellDB.DatabaseT,
                      module Hawl.Session.SessionT,
                      module Network.NewCGI.Trans
                     ) where

import Hawl.HaskellDB.DatabaseT
import Hawl.Session
import Hawl.Session.SessionT

import Network.NewCGI.Trans hiding (Html)

import Text.XHtml (URL)

import Control.Monad.Reader
import Data.Maybe (isJust)
import Network.URI

newtype Web c a = Web { unWeb :: ReaderT c (SessionT (DatabaseT (CGIT IO))) a }
    deriving (Functor, Monad, MonadDatabase,
              MonadSession, MonadCGI, MonadIO)

runWeb :: Database -> c -> Session -> Web c a -> CGI (a, Session)
runWeb db conf sess = flip runDatabaseT db . 
                      flip runSessionT sess .                        
                      flip runReaderT conf .  
                      unWeb

runWebSession :: Database -> c 
              -> URI -- ^  Session cookie URI
              -> Web c a -> CGI a
runWebSession db conf base x = 
    do sess <- loadOrStartSession db base
       (r,sess') <- runWeb db conf sess x
       when (sess' /= sess) $ updateSession db sess'
       return r

--
-- * Configuration
--

class BaseConfig c where
    baseURI :: c -> URI

instance BaseConfig c => BaseConfig (c,a) where
    baseURI = baseURI . fst

getConfig :: Web c c
getConfig = Web ask

fromConfig :: (c -> a) -> Web c a
fromConfig f = liftM f getConfig

withConfig :: (d -> c) -> Web c a -> Web d a
withConfig f w = do conf <- getConfig
                    Web $ lift $ flip runReaderT (f conf) $ unWeb w

configBaseURI :: BaseConfig c => Web c URI
configBaseURI = liftM baseURI getConfig

configBaseURL :: BaseConfig c => Web c URL
configBaseURL = liftM show configBaseURI

--
-- * Session
--

sessionCookieName :: String
sessionCookieName = "session"

loadSession :: (MonadIO m, MonadCGI m) => Database -> m (Maybe Session)
loadSession db = 
    do msi <- readCookie sessionCookieName
       maybe (return Nothing) (getSession db) msi

startSession :: (MonadIO m, MonadCGI m) => 
                Database 
             -> URI -- ^ Base URI for the session cookie.
             -> m Session
startSession db base =
    do s <- newSession db Nothing
       let c = Cookie { cookieName = sessionCookieName,
                        cookieValue = show (sessionID s),
                        cookieExpires = Nothing,
-- FIXME: this doesn't seem to work when the host is localhost
--                        cookieDomain = fmap uriRegName (uriAuthority base),
                        cookieDomain = Nothing,
                        cookiePath = Just $ uriPath base,
                        cookieSecure = False }
       setCookie c
       return s

loadOrStartSession :: (MonadIO m, MonadCGI m) => 
                      Database 
                   -> URI -- ^ Base URI for the session cookie.
                   -> m Session
loadOrStartSession db base = 
    loadSession db >>= maybe (startSession db base) return

endSession :: (MonadIO m, MonadCGI m) => 
              Database 
           -> URI -- ^ Base URI for the session cookie.
           -> Session -> m Session
endSession db base sess = 
    do deleteSession db $ sessionID sess
       startSession db base

--
-- * User
--

isLoggedIn :: MonadSession m => m Bool
isLoggedIn = liftM isJust currentSessionUser